#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <setjmp.h>
#include <signal.h>

#include "lib/log.h"
#include "wad/wadio.h"
#include "lib/timer.h"
#include "lib/conf.h"
#include "texture.h"
#include "gettable.h"
#include "plat/video.h"
#include "render/render.h"
#include "render/draw.h"
#include "banner.h"
#include "things.h"
#include "updmap.h"
#include "dsound.h"
#include "linetype.h"
#include "levinfo.h"
#include "plat/input.h"
#include "plat/sound.h"

/*#define WATCHDOG*/

#define DUMB "DUMB"
#define VERSION "0.07"

#ifdef __cplusplus
#define BANNER (DUMB "++ " VERSION)
#else
#define BANNER (DUMB " " VERSION)
#endif

#ifdef WATCHDOG
static jmp_buf alarm_jb;
static void alarm_handler(int i) {
   logprintf(LOG_ERROR,'D',
	     "got SIGALRM: renderer must have gotten stuck...");
   longjmp(alarm_jb,1);
};
#endif 

ConfItem mainconf[]={
   CONFB("auto-save-cfg",NULL,0,"save configuration after every run"),
   CONFNS("save-cfg",NULL,'S',"save configuration this time only"),
   CONFL("wad",NULL,'w',"specify wadfiles to load"),
   CONFS("map",NULL,'m',"set starting map","E1M1",9),
   CONFB("quiet",NULL,'q',"don't log to the screen"),
   CONFB("preview",NULL,0,"don't create any objects, just preview the map"),
   CONFI("difficulty",NULL,0,"set difficulty level",3),
   CONFB("silent",NULL,'s',"disable sound"),
   CONFI("width",NULL,0,"request width from video driver",320), 
   CONFI("height",NULL,0,"request height from video driver",200), 
   CONFI("depth",NULL,0,"request depth from video driver (in bytes/pixel)",1), 
   CONFS("log",NULL,'l',"log to this file",NULL,255),
   CONFS("record",NULL,0,"record player actions to file",NULL,255),
   CONFS("play",NULL,0,"playback player actions from file",NULL,255),
   CONFB("uncrowd",NULL,0,"hide gettables and banners"),
   CONFB("crosshair",NULL,0,"show a crosshair in the center of the view"),
   {NULL}
};
#define cnf_auto_save (mainconf[0].intval)
#define cnf_save (mainconf[1].intval)
#define cnf_wad (mainconf[2].listval)
#define cnf_map (mainconf[3].strval)
#define cnf_quiet (mainconf[4].intval)
#define cnf_preview (mainconf[5].intval)
#define cnf_difficulty (mainconf[6].intval)
#define cnf_silent (mainconf[7].intval)
#define cnf_width (mainconf[8].intval)
#define cnf_height (mainconf[9].intval)
#define cnf_bpp (mainconf[10].intval)
#define cnf_log (mainconf[11].strval)
#define cnf_record (mainconf[12].strval)
#define cnf_play (mainconf[13].strval)
#define cnf_uncrowd (mainconf[14].intval)
#define cnf_xhair (mainconf[15].intval)

ConfModule dumbconf[]={
   {input_conf,"input","Input Driver"},
   {video_conf,"vid","Video Driver"},
   {mainconf,"dumb","General Game Control"},
   {NULL,NULL,NULL}
};

/* this should all get moved into game.c */
static int width=0,height=0,bpp=0;
static int banner=-1,wbanner=-1,bfont=-1;
void gmsg(const char *s) {
   add_to_banner(banner,NULL,width,0);
   add_str_to_banner(banner,bfont,s);
};
static int want_new_lvl=0;
void game_want_newlvl(int x) {
   want_new_lvl=1+x;
};

/* control how much we're allowed to speed up the action */
#define MAXTICKS (500/MSEC_PER_TICK)

/* how often to check whether monsters wake up */
#define WAKE_TICKS (1000/MSEC_PER_TICK)

int main(int argc,char **argv) {
   char conf_file[256];
   void *fb;
   View view;
   LevData ld[1];
   ThingDyn *td;
   const SectorDyn *sd;
   int tt,frames=0;
   int i;
   int follow,shframes=0;
   FILE *frec=NULL,*fplay=NULL;
   int want_sound=1;
   int tickspassed=1;
   int crowd=1,showgetts=1;
   int wake_ticks=0;
   int load_failed=1;
   Texture *txh=NULL;

   LumpNum sound_ln=BAD_LUMPNUM;
   const unsigned char *sound=NULL;
   LumpNum sound2_ln=BAD_LUMPNUM;
   const unsigned char *sound2=NULL;

   INIT_RENDERER_SPEC((*init_renderer))=NULL;
   RENDER_SPEC((*render))=NULL;
   
   video_preinit();

#ifdef __MSDOS__
   find_conf_file(conf_file,"dumb.cfg");
#else
   find_conf_file(conf_file,".dumbrc");
#endif
   load_failed=load_conf(dumbconf,conf_file);
   if(conf_args(dumbconf,argc,argv)) return 1;

   if(!cnf_quiet) {
      /*setlinebuf(stdout);*/ /* if stdout is a socket, we'll want this */
      log_stdout();
   };
   logprintf(LOG_BANNER,'D',BANNER);

   if(!load_failed) 
      logprintf(LOG_INFO,'D',"Read configuration from %s",conf_file);

   if(cnf_uncrowd) crowd=showgetts=0;
   if(cnf_silent) want_sound=0;
   if(cnf_log) log_file(cnf_log,LOG_ALL,NULL);
   if(cnf_record) frec=fopen(cnf_record,"wb");
   if(cnf_play) fplay=fopen(cnf_record,"rb");
   width=cnf_width;
   height=cnf_height;
   bpp=cnf_bpp;

   if(cnf_wad) {
      char **s=cnf_wad;
      init_iwad(*(s++));
      while(*s) init_pwad(*(s++));
   }
   else if(*cnf_map=='E'||*cnf_map=='e') {
     init_iwad("doom.wad");
     init_pwad("doom4dum.wad");
   }
   else {
     init_iwad("doom2.wad");
     init_pwad("doom4dum.wad");
   };
   init_wadhashing();

   if(want_sound) {
      init_sound(11025);
      init_dsound();
   };
   init_textures();
   init_linetypes();
   init_levinfo();
   load_level(ld,cnf_map,cnf_difficulty,0);
   init_video(&width,&height,&bpp);
   video_winstuff(cnf_map,width,height);
   cnf_width=width;
   cnf_height=height;
   cnf_bpp=bpp;
   switch(bpp) {
#ifdef WANT_1BPP
   case(1):
      init_renderer=init_renderer8;
      render=render8;
      break;
#endif
#ifdef WANT_2BPP
   case(2):
      init_renderer=init_renderer16;
      render=render16;
      break;
#endif
#ifdef WANT_4BPP
   case(4):
      init_renderer=init_renderer32;
      render=render32;
      break;
#endif
   default:
      logfatal('D',"Unsupported BPP=%d",bpp);
   };
   init_input();
   set_playpal(0,video_setpal);

   if(cnf_preview) for(i=0;i<ldnthings(ld);i++) 
     if(i!=ld->player[0]) ldthingd(ld)[i].proto=NULL;
   
   td=ldthingd(ld)+(follow=ld->player[0]);

   init_view(&view);
   view.angle=0;
   view.x=td->x;
   view.y=td->y;
   if(td->sector==-1) 
     logprintf(LOG_ERROR,'D',"Can't find sector for player");
   view.sector=td->sector;
   sd=ldsectord(ld)+td->sector;
   /*if(vpdiv>1)
     init_renderer(width-width/vpdiv,height-height/vpdiv,width,height);
   else*/
     init_renderer(width,height,width,height);
   init_draw(width,height,bpp);

   init_gettables();
   set_gettable_special(1,&td->hits);

   banner=init_banner(64,0,width,10);
   wbanner=init_banner(height-12,0,width,3);
   if(have_lump("FONTB01"))
      bfont=init_font("FONTB%02d",64,'!'-1);
   else 
      bfont=init_font("STCFN%03d",128,0);
   if(crowd&&have_lump("DUMBLOGO")) 
      add_to_banner(banner,get_misc_texture("DUMBLOGO"),width,64);
   if(crowd&&have_lump("WAFFLE")) {
      LumpNum w=getlump("WAFFLE");
      add_to_banner(wbanner,NULL,width,0);
      add_text_to_banner(wbanner,bfont,
			 (const char *)load_lump(w),
			 get_lump_len(w));
   };

   if(cnf_xhair) txh=get_font_texture(bfont,'+');

   /*
   if(testarg) {
      tdraw=get_misc_texture(testarg);
      logprintf(LOG_DEBUG,'D',"texture %s: real=%dx%d log=%dx%d",
		tdraw->name,tdraw->width,tdraw->height,
		1<<tdraw->log2width,1<<tdraw->log2height);
   };
   */

   /* fill all video pages with our startup screen */
   /*if(have_lump("TITLEPIC")&&have_lump("DUMBLOGO")) {
      int i;
      Texture *t1=get_misc_texture("TITLEPIC");
      Texture *t2=get_misc_texture("DUMBLOGO");
      for(i=0;i<4;i++) {
	 fb=video_newframe();
	 draw_center(fb,t1);
	 draw_center(fb,t2);
	 video_updateframe(fb);
      };
   };*/
   
   tt=time(NULL);
   init_timer();
#ifdef WATCHDOG
   if(!setjmp(alarm_jb)) {
      signal(SIGALRM,alarm_handler);
#endif
      while(1) {
	 PlayerInput in;
	 int player_alive=td->hits>0;
	 if(ldthingd(ld)[follow].proto==NULL) follow=ld->player[0];
	 thing_to_view(ld,follow,&view);
	 dsound_setview(&view);
	 render(fb=video_newframe(),ld,&view);
	 if(showgetts) draw_gettables(fb,width,height);
	 if(tickspassed) update_banners(fb,tickspassed);
	 if(txh) draw(fb,txh,(width-txh->width)/2,(height-txh->height)/2);
	 video_updateframe(fb);
	 if(want_sound) poll_sound();
	 if(fplay) {
	    if(feof(fplay)) {
	       logprintf(LOG_INFO,'D',"Demo ran out...");
	       fclose(fplay);
	       fplay=NULL;
	    }
	    else fread(&in,sizeof(in),1,fplay);
	 };
	 if(fplay==NULL) get_input(&in);
	 if(frec) {
	    fwrite(&in,sizeof(in),1,frec);
	    fflush(frec);
	 };
	 if(in.quit) break;

	 /* only allow movement if player isn't dead */
	 if(td->proto!=NULL&&td->hits>0) {
	    if(in.sideways) 
	       thingd_apply_sideways(td,
				     INT_TO_FIXED(in.sideways)/(2*UNIT_SPEED));
	    if(in.forward) 
	       thingd_apply_forward(td,
				    INT_TO_FIXED(in.forward)/(2*UNIT_SPEED));
	    if(in.rotate) 
	       thingd_apply_torque(td,
				   INT_TO_FIXED(in.rotate)/(10*UNIT_SPEED));
	    if(in.jump) 
	       thingd_apply_up(td,INT_TO_FIXED(in.jump)/UNIT_SPEED);
	    if(in.shoot) 
	       use_selection(0,ld,ld->player[0]);
	    if(in.action) 
	       thing_activate(ld,ld->player[0],3*FIXED_ONE);
	 };

	 if(want_new_lvl) {
	    levinfo_next(ld,want_new_lvl>1);
	    reset_local_gettables();
	    td=ldthingd(ld)+(follow=ld->player[0]);
	    set_gettable_special(1,&td->hits);
	    video_winstuff(ld->name,width,height);
	    want_new_lvl=0;
	    continue;
	 };

	 /* check for special functions */
	 if(shframes<=0) {
	    if(in.select[0]) {
	       shframes=4;
	       want_new_lvl=1;
	    };
	    if(in.select[9]) {
	       int x,y;
	       unsigned int i;
	       const unsigned char *f=(const unsigned char *)fb;
	       FILE *fout;
	       logprintf(LOG_INFO,'D',"saving snapshot...");
	       fout=fopen("snapshot.ppm","w");
	       fprintf(fout,"P6\n%d %d\n%d\n",width,height,bpp==2?63:255);
	       for(y=0;y<height;y++) for(x=0;x<width;x++) 
		  switch(bpp) {
		  case(4):
		     putc(f[0],fout);
		     putc(f[1],fout);
		     putc(f[2],fout);
		     f+=4;
		     break;
		  case(2):
		     i=f[0]+((unsigned int)(f[1])<<8);
		     putc(((i>>10)&0x3e)|1,fout);
		     putc((i>>5)&0x3f,fout);
		     putc(((i&0x1f)<<1)|1,fout);
		     f+=2;
		     break;
		  case(1):
		     putc(0,fout);
		     putc(0,fout);
		     putc(0,fout);
		     f++;
		     break;
		  };
	       fclose(fout);
	       logprintf(LOG_INFO,'D',"snapshot done");
	    };
	    if(in.select[1]) {
	       shframes=4;
	       thing_hurl(ld,ld->player[0],8001,0,td->elev,0,1,0);
	    };
	    if(in.select[2]||in.w_sel) {
	       shframes=4;
	       /*thing_hurl(ld,ld->player[0],8002,0,td->elev);*/
	       rotate_selection(0);
	    };
	    if(in.select[6]) {
	       SoundBalance b=FLAT_BALANCE;
	       shframes=4;
	       if(!LUMPNUM_OK(sound_ln)) {
		  sound_ln=getlump("DSWPNUP");
		  sound=(const unsigned char *)load_lump(sound_ln);
	       };
	       play_sound(sound+8,get_lump_len(sound_ln)-8,
			  b,11025);
	    };
	    if(in.select[7]) {
	       SoundBalance b=FLAT_BALANCE;
	       shframes=4;
	       if(!LUMPNUM_OK(sound2_ln)) {
		  sound2_ln=getlump("DSITEMUP");
		  sound2=(const unsigned char *)load_lump(sound2_ln);
	       };
	       play_sound(sound2+8,get_lump_len(sound2_ln)-8,
			  b,11025);
	    };
	    if(in.select[8]) {
	       shframes=4;
	       cheat_gettables();
	       gmsg("CHEAT!");
	    };
	 };

	 /* back to ordinary main loop */
	 player_apply_lookup(ld,ld->player[0],in.lookup/UNIT_SPEED);
	 tickspassed=read_timer();
	 if(tickspassed>MAXTICKS) { 
	    logprintf(LOG_DEBUG,'D',"%d ticks passed: machine too slow?",
		      tickspassed);
	    tickspassed=MAXTICKS;
	 };
	 wake_ticks+=tickspassed;
	 if(wake_ticks>=WAKE_TICKS) {
	    thing_wake_others(ld,ld->player[0]);
	    wake_ticks=0;
	 };
	 if(tickspassed>0) {
	    reset_timer();
	    update_things(ld,tickspassed);
	    update_map(ld,tickspassed);
	    update_gettables(tickspassed);
	 }
	 else
	    logprintf(LOG_DEBUG,'D',
		      "congratulations: your processor is too fast!");

	 if(player_alive&&td->hits<=0)
	    gmsg("YOU DIED.  TOO BAD...");

	 if(shframes>0) shframes--;
	 frames++;
      };
#ifdef WATCHDOG
   };
   signal(SIGALRM,SIG_IGN);
#endif
   tt=time(NULL)-tt;
   logprintf(LOG_INFO,'D',"%d frames, %d seconds, %f fps",
	     frames,tt,(double)frames/tt);
   
   /*
   tt=time(NULL);
   for(view.height=sd->floor+(8<<12);view.height<sd->ceiling-(8>>12);view.height+=1<<11) {
      view.angle+=1<<9;
      render(fb=video_newframe(),ld,&view);
      video_updateframe(fb);
      frames++;
   };
   tt=time(NULL)-tt;
   logprintf(LOG_INFO,'D',"%ld frames, %ld seconds, %f fps",frames,tt,(double)frames/tt);
   */
   
   /*sleep(2);*/
   reset_input();
   reset_video();
   if(want_sound) {
      reset_sound();
      reset_dsound();
   };
   reset_gettables();
   free_level(ld);
   reset_levinfo();
   reset_linetypes();
   reset_textures();
   reset_wad();
   if(fplay) fclose(fplay);
   if(frec) fclose(frec);
   if(cnf_save||cnf_auto_save) {
      logprintf(LOG_INFO,'D',"Saving config to %s",conf_file);
      save_conf(dumbconf,conf_file,DIRT_ARGS);
   };
   log_exit();
   return 0;
};










